/* 
 * Copyright 2016 The Regents of the University of California
 * All rights reserved.
 * 
 * This file is part of Spoofer.
 * 
 * Spoofer is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Spoofer is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Spoofer.  If not, see <http://www.gnu.org/licenses/>.
 */

/****************************************************************************
   Author:      Ken Keys
   Date:        $Date: 2016/12/09 01:58:32 $
   Description: Traceroute wrapper
****************************************************************************/

#include "spoof.h"
#include "prober.h"
#include "AppInfo.h"
#include "routetracer.h"

static const char cvsid[] ATR_USED = "$Id: routetracer.cc,v 1.5 2016/12/09 01:58:32 kkeys Exp $";

bool RouteTracer::initializing = true;
RouteTracer::Type RouteTracer::type = UNKNOWN;
const char *RouteTracer::trace_program = nullptr;

// Split up scamper output by destination
void RouteTracer::splitScamperOutput(std::vector<in_addr> dests)
{
    family_t family = AF_INET;
    text[result[0].length] = '\0';
    result[0].text = nullptr;
    result[0].length = 0;
    char *start, *next;
    char addrbuf[INET6_ADDRSTRLEN+1];
    struct in_addr addr;
    for (start = text; *start; start = next) {
	next = strstr(start+1, "traceroute from");
	if (!next)
	    for (next = start; *next; next++);
#define STRINGIFY(X) STRINGIFY2(X)
#define STRINGIFY2(X) #X
#define WIDTH STRINGIFY(INET6_ADDRSTRLEN)
	if (sscanf(start, "traceroute from %*" WIDTH "[0-9a-fA-F.:] "
	    "to %" WIDTH "[0-9a-fA-F.:]s\n",
	    addrbuf) < 1)
		continue;
	if (inet_pton(family, addrbuf, &addr) != 1)
	    continue;
	for (unsigned i = 0; i < result.size(); i++) {
	    if (memcmp(&dests[i], &addr, addrlen(family)) == 0) {
		result[i].text = start;
		result[i].length = safe_int<size_t>(next - start);
		break;
	    }
	}
    }
}

bool RouteTracer::checkProg(Type checktype)
{
    std::vector<in_addr> addr(1);
    inet_pton(AF_INET, "127.0.0.1", &addr[0]);
    type = checktype;
    RouteTracer *rt = RouteTracer::make();
    bool works = rt->run(nullptr, addr) == 0 &&
	strstr(rt->text, "to 127.0.0.1");
    if (!works) type = MISSING;
    delete rt;
    return works;
}

bool RouteTracer::initScamper(char pathsep, const char *name, const char *alternate)
{
#ifndef SCAMPER_DISABLED
    // Use scamper if possible
    std::vector<const char *> scamper_candidate;
#ifdef SCAMPER_PATH
    scamper_candidate.push_back(SCAMPER_PATH);
#endif
    const char *path = AppInfo::path();
    if (path) {
        static char buf[PATH_MAX];
        const char *end = strrchr(path, pathsep);
        if (!end) end = path;
        snprintf(buf, sizeof(buf), "%.*s%c%s", int(end - path), path, pathsep, name);
        scamper_candidate.push_back(buf);
    }
    if (alternate)
	scamper_candidate.push_back(alternate);
    for (unsigned i = 0; i < scamper_candidate.size(); i++) {
        trace_program = scamper_candidate[i];
        if (checkProg(SCAMPER)) {
            printf("traceroute: %s\n", trace_program);
            initializing = false;
            return true;
        }
    }
#endif // !SCAMPER_DISABLED
    return false;
}

void RouteTracer::initRun(std::vector<in_addr> &dests)
{
    text = new char[BIGBUF*dests.size()];
    maxlen = safe_int<size_t>((type == SCAMPER) ? BIGBUF*dests.size() : BIGBUF);
    size_t childCount = (type == SCAMPER) ? 1 : dests.size();
    result.resize(dests.size());

    for (unsigned i = 0; i < result.size(); i++) {
        result[i].length = 0;
        result[i].text = (type != SCAMPER || i==0) ? &text[i*BIGBUF] : nullptr;
    }

    for (unsigned i = 0; i < childCount; i++) {
	children.push_back(RouteTracer::Child::make(i, this, dests));
    }

    start_time = time(nullptr);
    too_long = start_time + TRACEROUTE_MAX_WAIT +
	safe_int<time_t>((childCount - 1) * traceroute_delay/1000);
}

void RouteTracer::Child::adjustScamperGoal(Tracker *tracker)
{
    if (this->rt->type == SCAMPER && latest_hop != current_goal) {
	if (tracker) tracker->incGoal(latest_hop - current_goal);
	goal += latest_hop - current_goal;
	current_goal = latest_hop;
    }
}

void RouteTracer::Child::parse(Tracker *tracker)
{
    for ( ; scanned < result->length + this->done; scanned++) {
	if (scanned < result->length && result->text[scanned] == '\n') {
	    // normal end of line
	} else if (this->done && scanned == result->length) {
	    scanned++; // end of unterminated final line
	} else
	    continue; // middle of a line

	const char *line = result->text + parsed;

	size_t len = scanned - parsed;

	if (this->rt->type == SCAMPER && strncmp(line, "traceroute from", sizeof("traceroute from") - 1) == 0) {
	    // Beginning of new scamper target output
	    adjustScamperGoal(tracker); // update goal of previous target
	    latest_hop = 0;
	    current_goal = TRACEROUTE_ITEM_COST;
	} else {
	    char *t;
	    long hop = strtol(line, &t, 10);
	    if (hop >= 1 && hop <= TRACEROUTE_ITEM_COST && isspace(*t)) {
		// found a line starting with a hop number
		latest_hop = static_cast<int>(hop);
		bool gotResponse = false;
		for (len -= (unsigned)(t - line); len >= 3; t++, len--) {
		    if ((gotResponse = (strncmp(t, " ms", 3) == 0)))
			break;
		}
		if (this->rt->type == SCAMPER && latest_hop > current_goal) {
		    // Oops, we adjusted goal before target was done.  Fix it.
		    adjustScamperGoal(tracker);
		}
		if (tracker) tracker->incTries();
		this->tries++;
		if (gotResponse) {
		    if (tracker) tracker->incSuccess();
		    this->gap = 0;
		} else {
		    if (tracker) tracker->incFail();
		    this->gap++;
		}
	    }
	}

	parsed = scanned + 1;
    }
    if (!this->done && this->rt->type == NATIVE && this->gap >= TRACEROUTE_GAP_LIMIT) {
#if DEBUG
	printf("gap limit reached [child: %d]\n", this->id);
#endif
	this->abort = true;
    }
    if (this->rt->type == SCAMPER) {
	// Assuming the scamper output for the target is not split across
	// multiple parse() calls, the target is done.  Update its goal.
	adjustScamperGoal(tracker);
    }
}
