/* 
 * Copyright 2004-2009 Rob Beverly
 * Copyright 2015-2016 The Regents of the University of California
 * All rights reserved.
 * 
 * This file is part of Spoofer.
 * 
 * Spoofer is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Spoofer is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Spoofer.  If not, see <http://www.gnu.org/licenses/>.
 */

/****************************************************************************
   Author:      Rob Beverly <rbeverly@mit.edu>
                Ken Keys
   Date:        $Date: 2016/12/10 00:41:38 $
   Description: WIN32-Specific traceroute wrapper
****************************************************************************/
#ifdef _WIN32

#include "spoof.h"
#include "prober.h"
#include "routetracer.h"
static const char cvsid[] ATR_USED = "$Id: win_routetracer.cc,v 1.10 2016/12/10 00:41:38 kkeys Exp $";

struct WinRouteTracer : public RouteTracer {
    public:
    struct Child : public RouteTracer::Child {
	PROCESS_INFORMATION piProcInfo;
	HANDLE hThread;
	HANDLE StdOut_Rd;
	HANDLE StdOut_Wr;
	Child(unsigned _id, RouteTracer *_rt, std::vector<in_addr> &dests) :
	    RouteTracer::Child(_id, _rt, dests),
	    piProcInfo(), hThread(), StdOut_Rd(), StdOut_Wr()
	    {}
	Child(const Child &) NO_METHOD;
	Child operator=(const Child &) NO_METHOD;
	static DWORD WINAPI run(LPVOID lpParam);
    };
    WinRouteTracer(const WinRouteTracer &) NO_METHOD;
    WinRouteTracer operator=(const WinRouteTracer &) NO_METHOD;
    bool initClass();
public:
    WinRouteTracer() : RouteTracer() { }
    ~WinRouteTracer() { }
    int run(Tracker *tracker, std::vector<in_addr> &dests);
};

RouteTracer *RouteTracer::make()
    { return new WinRouteTracer(); }
RouteTracer::Child *RouteTracer::Child::make(unsigned _id, RouteTracer *_rt, std::vector<in_addr> &dests)
    { return new WinRouteTracer::Child(_id, _rt, dests); }


/* thread which performs tracing work */
DWORD WINAPI WinRouteTracer::Child::run(LPVOID lpParam) {
    char tracecmd[BUFSIZE];
    family_t family = AF_INET;
    DWORD status = 1; // failure

    /* from: http://msdn.microsoft.com/en-us/library/ms682499.aspx */
    STARTUPINFO siStartInfo;
    SECURITY_ATTRIBUTES saAttr;
    DWORD dwRead;
    BOOL bSuccess = FALSE;

    WinRouteTracer::Child *child = (WinRouteTracer::Child*)lpParam;
    RouteTracer::Result *&result = child->result;
    char addrbuf[INET6_ADDRSTRLEN+1];
    if (child->rt->type == SCAMPER) {
	int offset = sprintf(tracecmd, "\"%s\" -c \"trace -P icmp-paris -q 1 -w 1\" -i", trace_program);
	for (unsigned i = 0; i < child->num_dests; i++) {
	    inet_ntop(family, &child->dest[i], addrbuf, sizeof(addrbuf));
	    offset += sprintf(tracecmd + offset, " %s", addrbuf);
	}
    } else {
	inet_ntop(family, child->dest, addrbuf, sizeof(addrbuf));
	sprintf(tracecmd, "\"%s\" -w 1000 -d %s", trace_program, addrbuf);
    }

    saAttr.nLength = sizeof(SECURITY_ATTRIBUTES);
    saAttr.bInheritHandle = TRUE;
    saAttr.lpSecurityDescriptor = nullptr;

    CreatePipe(&child->StdOut_Rd, &child->StdOut_Wr, &saAttr, 0);
    SetHandleInformation(child->StdOut_Rd, HANDLE_FLAG_INHERIT, 0);

    ZeroMemory(&siStartInfo, sizeof(STARTUPINFO));
    siStartInfo.cb = sizeof(STARTUPINFO);
    siStartInfo.hStdError = nullptr;
    siStartInfo.hStdOutput = child->StdOut_Wr;
    siStartInfo.dwFlags |= STARTF_USESTDHANDLES;

    Sleep(traceroute_delay * child->id); // so all children don't probe the same hop simultaneously
    if (!initializing || verbosity >= HIGH)
	printf(">> traceroute[%d]: %s\n", child->id, tracecmd);
    if (!CreateProcess(nullptr, tracecmd, nullptr, nullptr, TRUE, 0,
        nullptr, nullptr, &siStartInfo, &child->piProcInfo))
    {
	debug(HIGH, "CreateProcess %d: error %ld\n", child->id, GetLastError());
	CloseHandle(child->StdOut_Wr);
	goto done;
    }
    CloseHandle(child->StdOut_Wr);
    while (1) {
        if (child->abort || child->rt->abort) {
            debug(HIGH, "stopping [thread: %d]\n", child->id);
	    if (!TerminateProcess(child->piProcInfo.hProcess, 1))
		debug(HIGH, "TerminateProcess: child %d: error %ld\n", child->id, GetLastError());
            break;
        }

	if (result->length >= child->rt->maxlen - 1) { // buffer full
	    if (!TerminateProcess(child->piProcInfo.hProcess, 1))
		debug(HIGH, "TerminateProcess: child %d: error %ld\n", child->id, GetLastError());
	    break;
	}
	bSuccess = ReadFile(child->StdOut_Rd, 
	    result->text + result->length, child->rt->maxlen - result->length - 1, &dwRead, nullptr);
	if (!bSuccess) {
	    if (GetLastError() == ERROR_BROKEN_PIPE) // child closed pipe
		status = 0; // success
	    break;
	}
	if (dwRead > 0)
	    result->length += dwRead;
        // XXX printf("Thread: %d (%p) Read: [%lu] Total: [%d]\n",
	    // XXX child->id, child, dwRead, result->length);
	if (dwRead <= 0)
            Sleep(200);
    }
    result->text[result->length] = '\0';
done:
    if (!CloseHandle(child->piProcInfo.hProcess))
	debug(HIGH, "Close hProcess: child %d: error %ld\n", child->id, GetLastError());
    if (!CloseHandle(child->piProcInfo.hThread))
	debug(HIGH, "Close hThread: child %d: error %ld\n", child->id, GetLastError());
    debug(HIGH, "Thread %d exit.\n", child->id);
    if (!CloseHandle(child->StdOut_Rd))
	debug(HIGH, "CloseHandle: child %d StdOut_Rd: error %ld\n", child->id, GetLastError());
    child->done = true;
    return status; // not ExitThread()!
}

bool WinRouteTracer::initClass()
{
    // XXX TODO need mutex if RouteTracer is used from multiple sessions
    if (type != UNKNOWN)
	return type != MISSING;

    if (initScamper('\\', "scamper.exe", nullptr))
	return true;

    // Fall back to platform's traceroute
    trace_program = "tracert";
    if (checkProg(NATIVE)) {
	printf("trace program: %s\n", trace_program);
	initializing = false;
	return true;
    }

    printf("Could not find a working traceroute program\n");
    trace_program = nullptr;
    initializing = false;
    return false;
}

int WinRouteTracer::run(Tracker *tracker, std::vector<in_addr> &dests)
{
    if (!initClass()) return -1;
    initRun(dests);

    /* Start threads */
    for (unsigned i = 0; i < children.size(); i++) {
	Child *child = static_cast<Child*>(children[i]);
        child->hThread = CreateThread(nullptr, 0, WinRouteTracer::Child::run, 
            child, 0, nullptr);
    }
    
    /* Wait for threads to finish, or timeout */
    while (true) {
        Sleep(200);
	int children_running = 0;
	for (unsigned i = 0; i < children.size(); i++) {
	    Child *child = static_cast<Child*>(children[i]);
	    if (child->scanned < child->result->length + child->done)
		child->parse(tracker);
	    if (child->done) {
		if (tracker) tracker->incGoal(child->tries - child->goal);
		child->goal = child->tries;
	    } else {
		children_running++;
	    }
	}
	if (!children_running) break;
        // XXX debug(DEVELOP, "%d threads active\n", children_running);
	time_t now = time(nullptr);
	if (now > too_long) {
	    if (!abort) {
		printf("aborting.");
		abort = true;
	    } else if (now > too_long + 10) {
		// threads are still running 10s after we asked them to exit
		break;
	    }
        }
    }

    /* Shut down threads */
    Sleep(1000);
    for (unsigned i = 0; i < children.size(); i++) {
	Child *child = static_cast<Child*>(children[i]);
	if (WaitForSingleObject(child->hThread, 0) != WAIT_OBJECT_0) {
            debug(HIGH, "Thread %d still running. shut it down.\n", i);
	    TerminateThread(child->hThread, 666);
	    WaitForSingleObject(child->hThread, INFINITE);
	    debug(HIGH, "Thread %d terminated.\n", i);
	    child->done = true;
	}
        CloseHandle(child->hThread);
	child->parse(tracker); // parse any remaining text
	if (tracker && child->tries < child->goal) {
	    tracker->incGoal(child->tries - child->goal);
	    child->goal = child->tries;
	}
    }

    if (type == SCAMPER)
	splitScamperOutput(dests);
    return 0;
}

#endif // _WIN32
